# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List


def fully_connected_network_spec(
    agents_by_type: Dict[str, List[str]]
) -> Dict[str, List[str]]:
    """Creates network spec for fully connected agents by agent type"""
    network_spec: Dict[str, List[str]] = {}
    for agent_type, agents in agents_by_type.items():
        all_agents = agents
        for agent in agents:
            network_spec[agent] = all_agents
    return network_spec
